/* ARCADEmini assembler (unix)
 * version 1.2 date 2003-09-12
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <malloc.h>

#include "id_list.h"
#include "node.h"
#include "global.h"
#include "debug.h"

//anchor of the double linked list with nodes
st_node _node_list =
{
  //set some data to keep compiler happy
  //this data is never used
  .type = node_generic,
  .data = { .generic = { } },
  .create_column = 0,
  .create_line = 0,
  //this are the important elements to initialize
  //previous and next is the anchor again: list is empty
  ._p_prev = &_node_list,
  ._p_next = &_node_list
};

//none register node
// - pointer to this node is used to indicate
//   that no "REGISTER" was found before
// - so if this is the parameter in the "NEXT_REGISTER"-node
//   the next register has to be $0
st_node node_reg_none =
{
  .type = node_reg,
  .data =
  {
    .reg =
    { 
      .p_con = NULL,
      .reg_no_avail = 0,
      .reg_no = 0
    }
  },
  .create_column = 0,
  .create_line = 0,
  ._p_prev = &node_reg_none,
  ._p_next = &node_reg_none
};

#ifdef DEBUG
//debug function: convert identifier type to human readable string
char * node_dbg_type2str( et_node_type type )
{
  switch( type )
  {
    case node_generic: return "generic";
    case node_con: return "con";
    case node_con_ident: return "con_ident";
    case node_con_unary: return "con_unary";
    case node_con_binary: return "con_binary";
    case node_con_regdif: return "con_regdif";
    case node_con_bra: return "reg_bra";
    case node_con_sec: return "reg_sec";
    case node_con_ofs: return "reg_ofs";
    case node_reg_con: return "reg_con";
    case node_reg_ident: return "reg_ident";
    case node_reg_binary: return "reg_binary";
    case node_reg_bra: return "reg_bra";
    case node_addr_ident: return "addr_indent";
    case node_sec_addr: return "sec_addr";
    case node_ofs_addr: return "ofs_addr";
    case node_reg: return "reg";
    case node_next_reg: return "next_reg";
    case node_str: return "str";
    case node_cmd: return "cmd";
    case node_lab: return "lab";
    case node_list: return "list";
    default: return "???";
  }
}
#endif

//create a new node
//(this function is for internal use only)
static st_node * node_new( et_node_type type, unsigned int extra_space )
{
  st_node * p_node;

  //allocate node and additional extra space behind this structure in same memory block
  p_node = (st_node *)malloc( sizeof( st_node ) + extra_space );
  if( p_node == NULL )
    return NULL;

  //set type
  p_node->type = type;

  //remember current position in input file
  p_node->create_column = column;
  p_node->create_line = line;

  //put node into node list
  p_node->_p_prev = &_node_list;
  p_node->_p_next = _node_list._p_next;
  p_node->_p_prev->_p_next = p_node;
  p_node->_p_next->_p_prev = p_node;

  //return pointer to node
  return p_node;
}

//create a new generic node
st_node * node_new_generic( )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_generic, 0 );
  if( p_node == NULL )
    return NULL;

  debug( "node_new_generic" );

  //return pointer to node
  return p_node;
}

//create a new constant node
st_node * node_new_con( unsigned long val )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con.val = val;

  debug1( "node_new_con: val=%lu", val );

  //return pointer to node
  return p_node;
}

//create a new constant identifier node
st_node * node_new_con_ident( st_id_list_entry * p_id )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_ident, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_ident.p_id = p_id;
  p_node->data.con_ident.val_avail = 0;
  p_node->data.con_ident.val = 0;

  debug1( "node_new_con_ident: id=%s", p_id->p_name );

  //return pointer to node
  return p_node;
}

//create a new unaray constant operator node
st_node * node_new_con_unary( char op, st_node * p_param )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_unary, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_unary.op = op;
  p_node->data.con_unary.p_param = p_param;
  p_node->data.con_unary.val_avail = 0;
  p_node->data.con_unary.val = 0;

  debug1( "node_new_con_unary: op=%c", op );

  //return pointer to node
  return p_node;
}

//create a new binary constant operator node
st_node * node_new_con_binary( char op, st_node * p_param1, st_node * p_param2 )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_binary, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_binary.op = op;
  p_node->data.con_binary.p_param1 = p_param1;
  p_node->data.con_binary.p_param2 = p_param2;
  p_node->data.con_binary.val_avail = 0;
  p_node->data.con_binary.val = 0;

  debug1( "node_new_con_binary: op=%c", op );

  //return pointer to node
  return p_node;
}

//create a new register-difference constant operator node
st_node * node_new_con_regdif( st_node * p_param1, st_node * p_param2 )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_regdif, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_regdif.p_param1 = p_param1;
  p_node->data.con_regdif.p_param2 = p_param2;
  p_node->data.con_regdif.val_avail = 0;
  p_node->data.con_regdif.val = 0;

  debug( "node_new_con_regdif" );

  //return pointer to node
  return p_node;
}

//create a new constant brackets node
st_node * node_new_con_bra( st_node * p_param )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_bra, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_bra.p_param = p_param;
  p_node->data.con_bra.val_avail = 0;
  p_node->data.con_bra.val = 0;

  debug( "node_new_con_bra" );

  //return pointer to node
  return p_node;
}

//create a new constant sector node
st_node * node_new_con_sec( st_node * p_addr )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_sec, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_sec.p_addr = p_addr;
  p_node->data.con_sec.val_avail = 0;
  p_node->data.con_sec.val = 0;

  debug( "node_new_con_sec" );

  //return pointer to node
  return p_node;
}

//create a new constant offset node
st_node * node_new_con_ofs( st_node * p_addr )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_con_ofs, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.con_ofs.p_addr = p_addr;
  p_node->data.con_ofs.val_avail = 0;
  p_node->data.con_ofs.val = 0;

  debug( "node_new_con_ofs" );

  //return pointer to node
  return p_node;
}

//create a new register constant node
st_node * node_new_reg_con( st_node * p_con )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_reg_con, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.reg_con.p_con = p_con;
  p_node->data.reg_con.reg_no_avail = 0;
  p_node->data.reg_con.reg_no = 0;

  debug( "node_new_reg_con" );

  //return pointer to node
  return p_node;
}

//create a new register identifier node
st_node * node_new_reg_ident( st_id_list_entry * p_id )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_reg_ident, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.reg_ident.p_id = p_id;
  p_node->data.reg_ident.reg_no_avail = 0;
  p_node->data.reg_ident.reg_no = 0;

  debug1( "node_new_reg_ident: id=%s", p_id->p_name );

  //return pointer to node
  return p_node;
}

//create a new binary register operator node
st_node * node_new_reg_binary( char op, st_node * p_reg, st_node * p_con )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_reg_binary, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.reg_binary.op = op;
  p_node->data.reg_binary.p_reg = p_reg;
  p_node->data.reg_binary.p_con = p_con;
  p_node->data.reg_binary.reg_no_avail = 0;
  p_node->data.reg_binary.reg_no = 0;

  debug1( "node_new_reg_binary: op=%c", op );

  //return pointer to node
  return p_node;
}

//create a new register brackets node
st_node * node_new_reg_bra( st_node * p_param )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_reg_bra, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.reg_bra.p_param = p_param;
  p_node->data.reg_bra.reg_no_avail = 0;
  p_node->data.reg_bra.reg_no = 0;

  debug( "node_new_reg_bra" );

  //return pointer to node
  return p_node;
}

//create a new address identifier node
st_node * node_new_addr_ident( st_id_list_entry * p_id )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_addr_ident, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.addr_ident.p_id = p_id;
  p_node->data.addr_ident.addr_avail = 0;
  p_node->data.addr_ident.addr_sec = 0;
  p_node->data.addr_ident.addr_ofs = 0;

  debug1( "node_new_addr_ident: id=%s", p_id->p_name );

  //return pointer to node
  return p_node;
}

//create a new sector address node
st_node * node_new_sec_addr( st_node * p_addr )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_sec_addr, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.sec_addr.p_addr = p_addr;
  p_node->data.sec_addr.val_avail = 0;
  p_node->data.sec_addr.val = 0;

  debug( "node_new_sec_addr" );

  //return pointer to node
  return p_node;
}

//create a new offset address node
st_node * node_new_ofs_addr( st_node * p_addr, unsigned long cur_addr_sec )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_ofs_addr, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.ofs_addr.p_addr = p_addr;
  p_node->data.ofs_addr.addr_sec = cur_addr_sec;
  p_node->data.ofs_addr.val_avail = 0;
  p_node->data.ofs_addr.val = 0;

  debug( "node_new_ofs_addr" );

  //return pointer to node
  return p_node;
}

//create a new register node ("REGISTER")
st_node * node_new_reg( st_node * p_con )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_reg, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.reg.p_con = p_con;
  p_node->data.reg.reg_no_avail = 0;
  p_node->data.reg.reg_no = 0;

  debug( "node_new_reg" );

  //return pointer to node
  return p_node;
}

//create a new next register node ("NEXT_REG")
st_node * node_new_next_reg( st_node * p_prev_reg )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_next_reg, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.next_reg.p_prev_reg = p_prev_reg;
  p_node->data.next_reg.reg_no_avail = 0;
  p_node->data.next_reg.reg_no = 0;

  debug( "node_new_next_reg" );

  //return pointer to node
  return p_node;
}

//create a new string node
st_node * node_new_str( unsigned char len )
{
  st_node * p_node;

  //create node
  //(space for data is also allocated within this memory block)
  p_node = node_new( node_str, len );
  if( p_node == NULL )
    return NULL;

  //set length of string
  p_node->data.str.len = len;
  //set pointer to buffer for string to point behind the node structure
  //(memory block was allocated large enough)
  p_node->data.str.data = (unsigned char *)p_node + sizeof( st_node );

  debug1( "node_new_str: len=%u", len );

  //return pointer to node
  return p_node;
}

//create a new command node
st_node * node_new_cmd( unsigned char param_cnt, unsigned char len )
{
  st_node * p_node;

  //create node
  //(space for parameters and data is also allocated within this memory block)
  p_node = node_new( node_cmd, param_cnt * sizeof( st_node_cmd_param ) + len );
  if( p_node == NULL )
    return NULL;

  //set number of parameters
  p_node->data.cmd.param_cnt = param_cnt;
  //set pointer to parameter array to point behind the node sturcture
  //(memory block was allocated large enough)
  p_node->data.cmd.params = (st_node_cmd_param *)((char *)p_node + sizeof( st_node ));

  //set length of command
  p_node->data.cmd.len = len;
  //set pointer to buffer for command data to point behind the parameters
  //(memory block was allocated large enough)
  p_node->data.cmd.data = (unsigned char *)p_node + sizeof( st_node ) + param_cnt * sizeof( st_node_cmd_param );
  //data is not yet available
  p_node->data.cmd.data_ok = 0;

  //set rest of data
  p_node->data.cmd.addr_sec = 0;
  p_node->data.cmd.addr_ofs = 0;

  debug2( "node_new_cmd: param_cnt=%u, len=%u", param_cnt, len );

  //return pointer to node
  return p_node;
}

//create a new label node
st_node * node_new_lab( st_id_list_entry * p_label, st_node * p_real_node )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_lab, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.lab.p_label = p_label;
  p_node->data.lab.p_real_node = p_real_node;

  debug( "node_new_lab" );

  //return pointer to node
  return p_node;
}

//create a new list node
st_node * node_new_list( st_node * p_entry, st_node * p_list )
{
  st_node * p_node;

  //create node
  p_node = node_new( node_list, 0 );
  if( p_node == NULL )
    return NULL;

  //set data
  p_node->data.list.p_entry = p_entry;
  p_node->data.list.p_list = p_list;

  debug( "node_new_list" );

  //return pointer to node
  return p_node;
}

//free a node
void node_free( st_node * p_node )
{
  //no node to free
  if( p_node == NULL )
    return;

  debug1( "node_free: type=%s", node_dbg_type2str( p_node->type ) );

  //remove node from node list
  p_node->_p_next->_p_prev = p_node->_p_prev;
  p_node->_p_prev->_p_next = p_node->_p_next;

  //free node
  free( p_node );
}

//free all nodes
void node_free_all( )
{
  debug( "node_free_all" );

  while( _node_list._p_next != &_node_list ) //until list is empty (anchor points to itself)
    node_free( _node_list._p_next ); //free first node in list
}

